import os
import pipes
import sys
import urllib
from io import StringIO
from itertools import product

import requests

from .keylib import KeyNotFound


class SedgeException(Exception):
    pass


class SecurityException(SedgeException):
    pass


class ParserException(SedgeException):
    pass


class OutputException(SedgeException):
    pass


class Section:
    def __init__(self, name, with_exprs):
        self.name = name
        self.with_exprs = with_exprs
        self.lines = []
        self.types = []
        self.identities = []

    def has_lines(self):
        return len(self.lines) > 0

    def add_line(self, keyword, parts):
        self.lines.append((keyword, parts))

    def add_type(self, name):
        self.types.append(name)

    def add_identity(self, name):
        self.identities.append(name)

    def get_lines(self, config_access, visited_set):
        """
        get the lines for this section
        visited_set is used to avoid visiting same section
        twice, if we've got a diamond in the @is setup
        """
        if self in visited_set:
            return []
        lines = self.lines.copy()
        visited_set.add(self)
        for identity in self.identities:
            if config_access.get_keyfile(identity):
                lines.append(('IdentitiesOnly', ['yes']))
                lines.append(('IdentityFile', [pipes.quote(config_access.get_keyfile(identity))]))
        for section_name in self.types:
            section = config_access.get_section(section_name)
            lines += section.get_lines(config_access, visited_set)
        return lines

    def __repr__(self):
        s = "{_type}:{name}".format(_type=type(self).__name__, name=self.name)
        if self.with_exprs:
            s += '[' + ','.join(' '.join(t) for t in self.with_exprs) + ']'
        return '<%s>' % s


class Root(Section):
    def __init__(self):
        super(Root, self).__init__('Root', [])
        self.pending_with = []
        self.vals = {}

    def add_type(self, name):
        raise ParserException('Cannot set an @is type on root scope.')

    def add_pending_with(self, parts):
        self.pending_with.append(parts)

    def pop_pending_with(self):
        r = self.pending_with
        self.pending_with = []
        return r

    def has_pending_with(self):
        return len(self.pending_with) > 0

    def set_value(self, key, value):
        self.vals[key] = value

    def get_variables(self):
        return self.vals

    def output_lines(self):
        for keyword, parts in sorted(self.lines):
            yield ConfigOutput.to_line(keyword, parts)


class HostAttrs(Section):
    def __init__(self, name):
        super(HostAttrs, self).__init__(name, [])


class Host(Section):
    @classmethod
    def expand_with_token(cls, s):
        fmt_error = 'range should be format {A..B} or {A..B/C}'
        if not s.startswith('{') or not s.endswith('}'):
            return [s]
        try:
            range_defn = s[1:-1]
            incr = 1
            if '/' in range_defn:
                range_defn, incr = range_defn.rsplit('/', 1)
                incr = int(incr)
            range_parts = range_defn.split('..')
            if len(range_parts) != 2:
                raise ParserException(fmt_error)
            from_val, to_val = (int(t) for t in range_parts)
            to_val += 1  # inclusive end
            from_width = len('%0s' % range_parts[0])
            to_width = len('%0s' % range_parts[1])
        except ValueError:
            raise ParserException('expected an integer in range definition.')
        if from_width == to_width:
            return ["%0*d" % (to_width, t) for t in range(from_val, to_val, incr)]
        else:
            return list(str(t) for t in range(from_val, to_val, incr))

    @classmethod
    def expand_with(cls, defn):
        expanded = []
        for tok in defn:
            expanded += Host.expand_with_token(tok)
        return expanded

    def resolve_defn(self, config_access):
        visited = set()
        # we shove the name in here, then on the other end of the substitution
        # logic we can get it back out. FIXME clean this up.
        lines = [self.name]
        for keyword, parts in self.get_lines(config_access, visited):
            lines.append(ConfigOutput.to_line(keyword, parts, indent=4))
        return lines

    def apply_substitutions(self, lines, val_dict):
        for line in lines:
            for subst, value in val_dict.items():
                line = line.replace(subst, value)
            yield line

    def variable_iter(self, base):
        """
        returns iterator over the cross product of the variables
        for this stanza
        """
        base_substs = dict(('<' + t + '>', u) for (t, u) in base.items())
        substs = []
        vals = []
        for with_defn in self.with_exprs:
            substs.append('<' + with_defn[0] + '>')
            vals.append(Host.expand_with(with_defn[1:]))
        for val_tpl in product(*vals):
            r = base_substs.copy()
            r.update(dict(zip(substs, val_tpl)))
            yield r

    def host_stanzas(self, config_access):
        """
        returns a list of host definitions
        """
        defn_lines = self.resolve_defn(config_access)
        for val_dict in self.variable_iter(config_access.get_variables()):
            subst = list(self.apply_substitutions(defn_lines, val_dict))
            host = subst[0]
            lines = [ConfigOutput.to_line('Host', [host])] + subst[1:]
            yield host, lines


class SectionConfigAccess:
    """
    sections may require access to other parts of the file.
    this class provides that access.
    """

    def __init__(self, config):
        self._config = config

    def get_section(self, name):
        return self._config._get_section_by_name(name)

    def get_keyfile(self, name):
        try:
            fingerprints = self._config.keydefs[name]
        except KeyError:
            self._config.warn("identity '{}' is not defined (missing @key definition)".format(name))
            return None
        for fingerprint in fingerprints:
            try:
                return self._config._key_library.lookup(fingerprint)
            except KeyNotFound:
                pass
        self._config.warn("identity '{name}' (fingerprints {fingerprints}) not found in SSH key library".format(
            name=name,
            fingerprints='; '.join(fingerprints))
        )

    def get_variables(self):
        return self._config.sections[0].get_variables()


class ConfigOutput:
    def __init__(self, fd):
        self._fd = fd
        self.need_break = False

    def write_stanza(self, it):
        if self.need_break:
            self._fd.write('\n')
        for i, line in enumerate(it):
            if i == 0:
                self.need_break = True
            self._fd.write(line + '\n')

    @classmethod
    def to_line(cls, keyword, parts, indent=0):
        def add_indent(s):
            return ' ' * indent + s

        if len(parts) == 1:
            return add_indent(' '.join([keyword, '=', parts[0]]))
        out = [keyword]
        for part in parts:
            if '"' in part:
                raise OutputException("quotation marks may not be used in arguments")
            if ' ' in part:
                out.append("{}".format(part))
            else:
                out.append(part)
        return add_indent(' '.join(out))


class SedgeEngine:
    """
    base parser for a sedge configuration file.
    handles all directives and expansions
    """

    def __init__(self, key_library, fd, verify_ssl, url=None, args=None, parent_keydefs=None, via_include=False):
        self._key_library = key_library
        self._url = url
        self._args = args
        self._verify_ssl = verify_ssl
        self._via_include = via_include
        self.sections = [Root()]
        self.includes = []
        self.keydefs = {}
        if parent_keydefs is not None:
            self.keydefs.update(parent_keydefs)
        self.parse(fd)

    def warn(self, message):
        print("{url}: {msg}".format(url=self._url, msg=message), file=sys.stderr)

    @classmethod
    def parse_other_space(cls, other):
        in_quote = False
        args = []
        current = []

        def pop_current():
            val = ''.join(current)
            if val:
                args.append(val)
            current.clear()

        for c in other:
            if in_quote:
                if c == '"':
                    in_quote = False
                    pop_current()
                else:
                    current.append(c)
            else:
                if c == '"':
                    if len(current) > 0:
                        raise ParserException('quotation marks cannot be used within an argument value')
                    in_quote = True
                else:
                    if c == ' ':
                        pop_current()
                    else:
                        current.append(c)
        if in_quote:
            raise ParserException('unterminated quotation marks')
        pop_current()
        return args

    # from the ssh_config manual page:
    #  > ... format ``keyword arguments''.  Configuration options may be
    #  > separated by whitespace or optional whitespace and exactly one `='; the
    #  > latter format is useful to avoid the need to quote whitespace when speci-
    #  > fying configuration options using the ssh, scp, and sftp -o option.
    #  > Arguments may optionally be enclosed in double quotes (") in order to
    #  > represent arguments containing spaces.
    @classmethod
    def parse_config_line(cls, line):
        if '=' in line:
            line_parts = line.strip().split('=', 1)
            return line_parts[0].rstrip(), [line_parts[1].lstrip()]
        else:
            line_parts = line.strip().split(' ', 1)
            other = ''
            if len(line_parts) == 2:
                other = line_parts[1].strip()
            return line_parts[0], SedgeEngine.parse_other_space(other)

    def is_include(self):
        return self._via_include

    def parse(self, fd):
        """very simple parser - but why would we want it to be complex?"""

        def resolve_args(args):
            # FIXME break this out, it's in common with the templating stuff elsewhere
            root = self.sections[0]
            val_dict = dict(('<' + t + '>', u) for (t, u) in root.get_variables().items())
            resolved_args = []
            for arg in args:
                for subst, value in val_dict.items():
                    arg = arg.replace(subst, value)
                resolved_args.append(arg)
            return resolved_args

        def handle_section_defn(keyword, parts):
            if keyword == '@HostAttrs':
                if len(parts) != 1:
                    raise ParserException('usage: @HostAttrs <hostname>')
                if self.sections[0].has_pending_with():
                    raise ParserException('@with not supported with @HostAttrs')
                self.sections.append(HostAttrs(parts[0]))
                return True
            if keyword == 'Host':
                if len(parts) != 1:
                    raise ParserException('usage: Host <hostname>')
                self.sections.append(Host(parts[0], self.sections[0].pop_pending_with()))
                return True

        def handle_vardef(root, keyword, parts):
            if keyword == '@with':
                root.add_pending_with(parts)
                return True

        def handle_set_args(_, parts):
            if len(parts) == 0:
                raise ParserException('usage: @args arg-name ...')
            if not self.is_include():
                return
            if self._args is None or len(self._args) != len(parts):
                raise ParserException('required arguments not passed to include {url} ({parts})'.format(
                    url=self._url,
                    parts=', '.join(parts))
                )
            root = self.sections[0]
            for key, value in zip(parts, self._args):
                root.set_value(key, value)

        def handle_set_value(_, parts):
            if len(parts) != 2:
                raise ParserException('usage: @set <key> <value>')
            root = self.sections[0]
            root.set_value(*resolve_args(parts))

        def handle_add_type(section, parts):
            if len(parts) != 1:
                raise ParserException('usage: @is <HostAttrName>')
            section.add_type(parts[0])

        def handle_via(section, parts):
            if len(parts) != 1:
                raise ParserException('usage: @via <Hostname>')
            section.add_line(
                'ProxyCommand',
                ('ssh {args} nc %h %p 2> /dev/null'.format(args=pipes.quote(resolve_args(parts)[0])), )
            )

        def handle_identity(section, parts):
            if len(parts) != 1:
                raise ParserException('usage: @identity <name>')
            section.add_identity(resolve_args(parts)[0])

        def handle_include(_, parts):
            if len(parts) == 0:
                raise ParserException('usage: @include <https://...|/path/to/file.sedge> [arg ...]')
            url = parts[0]
            parsed_url = urllib.parse.urlparse(url)
            if parsed_url.scheme == 'https':
                req = requests.get(url, verify=self._verify_ssl)
                text = req.text
            elif parsed_url.scheme == 'file':
                with open(parsed_url.path) as fd:
                    text = fd.read()
            elif parsed_url.scheme == '':
                path = os.path.expanduser(url)
                with open(path) as fd:
                    text = fd.read()
            else:
                raise SecurityException('error: @includes may only use paths or https:// or file:// URLs')

            subconfig = SedgeEngine(
                self._key_library,
                StringIO(text),
                self._verify_ssl,
                url=url,
                args=resolve_args(parts[1:]),
                parent_keydefs=self.keydefs,
                via_include=True)
            self.includes.append((url, subconfig))

        def handle_keydef(_, parts):
            if len(parts) < 2:
                raise ParserException('usage: @key <name> [fingerprint]...')
            name = parts[0]
            fingerprints = parts[1:]
            self.keydefs[name] = fingerprints

        def handle_keyword(section, keyword, parts):
            handlers = {
                '@set': handle_set_value,
                '@args': handle_set_args,
                '@is': handle_add_type,
                '@via': handle_via,
                '@include': handle_include,
                '@key': handle_keydef,
                '@identity': handle_identity
            }
            if keyword in handlers:
                handlers[keyword](section, parts)
                return True

        for line in (t.strip() for t in fd):
            if line.startswith('#') or line == '':
                continue
            keyword, parts = SedgeEngine.parse_config_line(line)
            if handle_section_defn(keyword, parts):
                continue
            if handle_vardef(self.sections[0], keyword, parts):
                continue
            current_section = self.sections[-1]
            if handle_keyword(current_section, keyword, parts):
                continue
            if keyword.startswith('@'):
                raise ParserException("unknown expansion keyword {}".format(keyword))
            # use other rather than parts to avoid messing up user
            # whitespace; we don't handle quotes in here as we don't
            # need to
            current_section.add_line(keyword, parts)

    def sections_for_cls(self, cls):
        return (t for t in self.sections if isinstance(t, cls))

    def _get_section_by_name(self, name):
        matches = [t for t in self.sections if t.name == name]
        if len(matches) > 1:
            raise ParserException("More than one section with name '{}'".format(name))
        if len(matches) == 0:
            raise ParserException("No such section: {}".format(name))
        return matches[0]

    def host_stanzas(self):
        for host in self.sections_for_cls(Host):
            for tpl in host.host_stanzas(SectionConfigAccess(self)):
                yield tpl

    def output(self, out, stanza_names=None):
        # output global config from root section
        root = self.sections[0]
        if self.is_include():
            if root.has_lines():
                print("Warning: global config in @include '{url}' ignored.".format(url=self._url), file=sys.stderr)
                print("Ignored lines are:", file=sys.stderr)
                warning_fd = StringIO()
                warning_out = ConfigOutput(warning_fd)
                warning_out.write_stanza(root.output_lines())
                print("\n".join([" > " + t for t in warning_fd.getvalue().splitlines()]), file=sys.stderr)
        else:
            out.write_stanza(root.output_lines())

        if stanza_names is None:
            stanza_names = set()

        dupes = set()
        for hostname, stanza in self.host_stanzas():
            if hostname in stanza_names:
                dupes.add(hostname)
            stanza_names.add(hostname)
            out.write_stanza(stanza)
        if dupes:
            print("Warning: duplicated hosts parsing '{url}'".format(url=self._url), file=sys.stderr)
            print("  %s" % (', '.join(sorted(dupes))), file=sys.stderr)

        for url, subconfig in self.includes:
            subconfig.output(out, stanza_names)

        # write out a list of hosts for completion use
        if not self.is_include():
            outf = os.path.expanduser("~/.sedge/hosts")
            pattern_characters = (',', '!', '*', '?')
            try:
                with open(outf, 'w') as fd:
                    for host in sorted(stanza_names):
                        if any(host.find(c) != -1 for c in pattern_characters):
                            continue
                        print(host, file=fd)
            except IOError:
                print("warning: ~/.sedge/hosts could not be written.", file=sys.stderr)



import os
import re
import subprocess
import sys


class KeyNotFound(Exception):
    pass


class FingerprintDoesNotParse(Exception):
    pass


class KeyLibrary:
    fingerprint_re = re.compile(r'^(?P<length>\d+) (?P<fingerprint>[^ ]+) (?P<comment>.*) (?P<algorithm>\(.*\)$)')

    def __init__(self, path, verbose=False):

        self._path = path
        self._verbose = verbose
        self.keys_by_fingerprint = {}
        self._scan()

    def _generate_public_key(self, fname):
        pkey_fname = fname + '.pub'
        if os.access(pkey_fname, os.R_OK):
            return
        print("public key does not exist for private key '{name}'".format(name=fname), file=sys.stderr)
        print("attempting to generate; you may be prompted for a pass phrase.", file=sys.stderr)
        try:
            public_key = subprocess.check_output(['ssh-keygen', '-y', '-f', fname])
        except subprocess.CalledProcessError:
            return
        with open(pkey_fname, 'wb') as fd:
            fd.write(public_key)
        print("Generated public key successfully.", file=sys.stderr)
        return True

    @classmethod
    def _fingerprint_from_keyinfo(cls, output):
        m = KeyLibrary.fingerprint_re.match(output)
        if not m:
            raise FingerprintDoesNotParse()
        return m.group('fingerprint')

    def _scan_key(self, fname, recurse=False):
        try:
            output = subprocess.check_output(['ssh-keygen', '-l', '-f', fname]).decode('utf8')
            try:
                return KeyLibrary._fingerprint_from_keyinfo(output)
            except FingerprintDoesNotParse:
                print("warning: public key fingerprint couldn't be parsed: '%s'" % fname, file=sys.stderr)
                print(output, file=sys.stderr)
        except subprocess.CalledProcessError:
            if not recurse and self._generate_public_key(fname):
                return self._scan_key(fname, recurse=True)

    def _scan(self):
        def rp(path):
            return os.path.relpath(path, self._path)

        skip = {'config', 'known_hosts', 'known_hosts.old', 'authorized_keys'}
        for dirpath, dirnames, fnames in os.walk(self._path):
            for name, path in ((t, os.path.join(dirpath, t)) for t in fnames):
                if name.startswith('.'):
                    continue
                if name.endswith('.pub'):
                    continue
                if name in skip:
                    continue
                fingerprint = self._scan_key(path)
                if fingerprint is not None:
                    if self._verbose:
                        print("scanned key '{key}' fingerprint '{fingerprint}'".format(
                            key=rp(path),
                            fingerprint=fingerprint)
                        )
                    if fingerprint in self.keys_by_fingerprint:
                        print(
                            "warning: key '{key}' has same fingerprint as '{otherkey}', ignoring duplicate key.".format(
                                key=rp(self.keys_by_fingerprint[fingerprint]),
                                otherkey=rp(path))
                        )
                    else:
                        self.keys_by_fingerprint[fingerprint] = path

    def list_keys(self):
        max_finger = max(len(t) for t in self.keys_by_fingerprint)
        for k, v in sorted(self.keys_by_fingerprint.items(), key=lambda x: x[1]):
            print("%*s  %s" % (max_finger, k, v))

    def add_keys(self):
        files = list(sorted(set(self.keys_by_fingerprint.values())))
        subprocess.call(['ssh-add'] + files)

    def lookup(self, fingerprint):
        try:
            return self.keys_by_fingerprint[fingerprint]
        except KeyError:
            raise KeyNotFound()



import difflib
import os.path
import sys
from pathlib import Path
from tempfile import NamedTemporaryFile

import click

from .engine import SedgeEngine, ConfigOutput
from .keylib import KeyLibrary
from .templates import sedge_config_header


def get_file_backup_name(file_name):
    return file_name + '.pre_sedge'


def ask_overwrite(file_name):
    click.echo("WARNING: '{}' already exists and was not generated by sedge.".format(file_name), err=True)
    confirm = input("Enter 'yes' to overwrite (I'll make a backup called '{backup_file}'): ".format(
        backup_file=get_file_backup_name(file_name))
    )
    return confirm == 'yes'


def backup_file(file_name):
    _backup_file = get_file_backup_name(file_name)
    os.rename(file_name, _backup_file)
    click.echo('Your previous SSH configuration file has been renamed to: {}'.format(_backup_file), err=True)


def check_or_confirm_overwrite(file_name):
    """
    Returns True if OK to proceed, False otherwise
    """
    try:
        with open(file_name) as fd:
            header = next(fd)
            if header.find(':sedge:') == -1:
                okay = ask_overwrite(file_name)
                if okay:
                    backup_file(file_name)
                else:
                    return False

    except FileNotFoundError:
        click.echo("{} not found".format(file_name), err=True)
    except StopIteration as e:
        click.echo(repr(e), err=True)
    else:
        return True


def diff_config_changes(before, after):
    def get_data(f):
        try:
            with open(f) as fd:
                return fd.read().splitlines(True)
        except FileNotFoundError:
            click.echo("{} not found".format(f), err=True)
            return []

    a = get_data(before)
    b = get_data(after)
    diff_lines = list(difflib.unified_diff(a, b))
    if not diff_lines:
        click.echo('no changes.', err=True)
    else:
        click.echo('configuration changes:', err=True)
        click.echo(''.join(diff_lines), err=True)


class SedgeConfig:
    pass


sedge_config = click.make_pass_decorator(SedgeConfig, ensure=True)


@click.group()
@click.version_option()
@click.option('-c',
              '--config-file',
              default=os.path.expanduser('~/.sedge/config'), )
@click.option('-o',
              '--output-file',
              default=os.path.expanduser('~/.ssh/config'), )
@click.option('-n',
              '--no-verify',
              is_flag=True,
              help='do not verify HTTPS requests')
@click.option('-k',
              '--key-directory',
              default=os.path.expanduser('~/.ssh'),
              help='directory to scan for SSH keys', )
@click.option('-v', '--verbose', count=True, default=0)
@sedge_config
def cli(config, verbose, key_directory, no_verify, output_file, config_file):
    """
    Template and share OpenSSH ssh_config(5) files. A preprocessor for
    OpenSSH configurations.
    """
    config.verbose = verbose
    config.key_directory = key_directory
    config.config_file = config_file
    config.output_file = output_file
    config.no_verify = no_verify


@cli.command('init')
@sedge_config
def init(config):
    """
    Initialise ~./sedge/config file if none exists.
    Good for first time sedge usage
    """
    from pkg_resources import resource_stream
    import shutil

    config_file = Path(config.config_file)
    if config_file.is_file():
        click.echo('{} already exists, maybe you want $ sedge update'.format(config_file))
        sys.exit()

    config_file.parent.mkdir(parents=True, exist_ok=True)
    with resource_stream(__name__, 'sedge_template.conf') as src_stream:
        with open(config.config_file, 'wb') as target_stream:
            shutil.copyfileobj(src_stream, target_stream)


@cli.command('update')
@sedge_config
def update(config):
    """
    Update ssh config from sedge specification
    """

    def write_to(out):
        engine.output(out)

    config_file = Path(config.config_file)
    if not config_file.is_file():
        click.echo('No file {} '.format(config_file), err=True)
        sys.exit()

    library = KeyLibrary(config.key_directory)
    with config_file.open() as fd:
        engine = SedgeEngine(library, fd, not config.no_verify, url=config.config_file)

    if config.output_file == '-':
        write_to(ConfigOutput(sys.stdout))
        return

    if not check_or_confirm_overwrite(config.output_file):
        click.echo('Aborting.', err=True)
        sys.exit(1)

    tmp_file = NamedTemporaryFile(mode='w', dir=os.path.dirname(config.output_file), delete=False)
    try:
        tmp_file.file.write(sedge_config_header.format(config.config_file))
        write_to(ConfigOutput(tmp_file.file))
        tmp_file.close()
        if config.verbose:
            diff_config_changes(config.output_file, tmp_file.name)
        os.rename(tmp_file.name, config.output_file)
    except:
        os.unlink(tmp_file.name)
        raise


@cli.group()
@sedge_config
def keys(config):
    """
    Manage ssh keys
    """


@keys.command('list')
@sedge_config
def command_list_keys(config):
    library = KeyLibrary(path=config.key_directory, verbose=config.verbose)
    library.list_keys()


@keys.command('add')
@sedge_config
def command_add_keys(config):
    library = KeyLibrary(path=config.key_directory, verbose=config.verbose)
    library.add_keys()



from setuptools import setup, find_packages

dev_requires = [
    'flake8',
    'pytest',
    'tox'
]

install_requires = [
    'requests',
    'click'
]

long_description = """
Template and share OpenSSH ssh_config(5) files.
A preprocessor for OpenSSH configurations.
Named for the favourite food of the Western Ground Parrot.
If you find this software useful, please consider donating to the effort to save this critically endangered species.

http://www.western-ground-parrot.org.au/
"""

setup(
    author="Grahame Bowland",
    author_email="grahame@angrygoats.net",
    description="Template and share OpenSSH ssh_config files.",
    long_description=long_description,
    license="GPL3",
    keywords="openssh ssh",
    url="https://github.com/grahame/sedge",
    name="sedge",
    version="2.0.3",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    extras_require={
        'dev': dev_requires
    },
    install_requires=install_requires,
    entry_points={
        'console_scripts': [
            'sedge = sedge.cli:cli',
        ],
    },
    package_data={'sedge': ['sedge_template.conf']},
    include_package_data=True,
)
